
/* Boolean formulas program (BOMO), C++
   Lutz Prechelt, 1997-11-07
   $Id: both.cc,v 1.2 1997/11/10 18:04:22 prechelt Exp prechelt $
*/

#include <stream.h>

// inheritance hierarchy (indentation means 'inherits from'):
 class ValuesOfVariables;
 class Formula;
    class NfoldTerm;
       class AndTerm;
       class OrTerm;
    class XorTerm;
    class NotTerm;
    class VarTerm;

// auxiliary procedure: begin a new line and ident as indicated:
static void indent(ostream *out, int indentation); 

//****************************************************************

/* assign values true or false to variables represented as Strings.
 */
class ValuesOfVariables {
private:
  StringSet *s;  // dynamic set of char* values
  // a variable is considered true  iff  its name is present in the set
  
public:
  ValuesOfVariables()
    {
      s = new StringSet();
    }
    
  virtual ~ValuesOfVariables()
    {
      delete s;
    }
    
  virtual void setTrue(char *variable)
    {
      s.insert(variable); // does nothing if s.contains(variable)
    }
    
  virtual void setFalse(char *variable)
    {
      s.remove(variable);  // does nothing if !s.contains(variable)
    }
    
  virtual bool isTrue(char *variable)
    {
      return s.contains(variable);
    }
};

//****************************************************************

/* represent any boolean formula.
   root class of a COMPOSITE.
 */
class Formula {
public:
  // return the n-th operand of top operator:
  virtual Formula* getOperand(int number) = 0;
  // return the number of operands of top operator (variables have 0):
  virtual int   numberOfOperands() = 0;
  // return the name of the top operator of this formula:
  virtual char* operatorname() = 0;
  // print the formula in multi-line, indented, prefix form:
  virtual void  prettyprint(ostream *out, int indentation = 0) = 0;
  // print the formula in one-line, parenthesized, infix form:
  virtual void  print(ostream *out) = 0;
};

//****************************************************************

/* abstract container class for COMPOSITE.
   used for operations that can have an arbitrary number of operands
 */
class NfoldTerm : public Formula {
private:
  const int maxNoOfOperands = 50; // max this many factors in one term
  Formula  *operand[maxNoOfOperands];  // factors of this term
  int       noOfOperands;              // how many factors there are

public:
  NfoldTerm()
    {
      noOfOperands = 0;
    }

  virtual ~NfoldTerm()
    {
      for (int i = 0; i < noOfOperands; i++) {
        delete operand[i];
      }
    }

  virtual int addOperand(Formula *newoperand)
    {
      if (noOfOperands < maxNoOfOperands)
        operand[noOfOperands++] = newoperand;
      return noOfOperands;
    }    

  virtual Formula *getOperand(int number)
    {
      return (number < noOfOperands) ? operand[number] : (Formula*)NULL;
    }
    
  virtual int numberOfOperands()
    {
      return noOfOperands;
    }

  virtual void  print(ostream *out)
    {
      *out << "(";
      for (int i = 0; i < noOfOperands; i++) {
        operand[i]->print(out);
        if (i+1 < noOfOperands) // if this is not the last operand
          *out << " " << operatorname() << " ";
      }
      *out << ")";
    }

  virtual void  prettyprint(ostream *out, int indentation = 0)
    {
      indent(out, indentation);
      *out << operatorname();
      for (int i = 0; i < noOfOperands; i++)
        operand[i]->prettyprint(out, indentation+2);
    }

};

//****************************************************************

/* class for terms of the form  a AND b AND c AND ...
   is a container class of a COMPOSITE
 */
class AndTerm : public NfoldTerm {
public:
  AndTerm()
    {
      // an empty AND
    }

  AndTerm(Formula *left, Formula *right) : NfoldTerm()
    {
      // enter two terms as initial factors (others may be added later):
      addOperand(left);
      addOperand(right);
    }

  virtual ~AndTerm()
    {
      // just call destructor of NfoldTerm
    }

  virtual char* operatorname()
    {
      return "AND";
    }
};

//****************************************************************

/* class for terms of the form  a OR b OR c OR ...
   is a container class of a COMPOSITE
 */
class OrTerm : public NfoldTerm {
public:
  OrTerm()
    {
      // an empty OR
    }

  OrTerm(Formula *left, Formula *right) : NfoldTerm()
    {
      // enter two terms as initial factors (others may be added later):
      addOperand(left);
      addOperand(right);
    }

  virtual ~OrTerm()
    {
      // just call destructor of NfoldTerm
    }

  virtual char* operatorname()
    {
      return "OR";
    }
};

//****************************************************************

/* class for terms of the form  a XOR b
   This is a leaf class of a COMPOSITE
 */
class XorTerm : public Formula {
private:
  Formula *left, *right;  // the operands

public:
  XorTerm(Formula *a, Formula *b)
    {
      left = a;
      right = b;
    }
    
  virtual ~XorTerm()
    {
      delete left;
      delete right;
    }

  virtual Formula* getOperand(int number)
    {
      switch(number) {
      case 0 : return left;
      case 1 : return right;
      default: return (Formula*)NULL;
      }
    }

  virtual int numberOfOperands()
    {
      return 2;
    }
    
  virtual char* operatorname()
    {
      return "XOR";
    }

  virtual void  print(ostream *out)
    {
      *out << "(";
      left->print(out);
      *out << " " << operatorname() << " ";
      right->print(out);
      *out << ")";
    }

  virtual void  prettyprint(ostream *out, int indentation = 0)
    {
      indent(out, indentation);
      *out << operatorname();
      left->prettyprint(out, indentation+2);
      right->prettyprint(out, indentation+2);
    }
};

//****************************************************************

/* class for terms of the form  NOT a.
   This is a leaf class of a COMPOSITE
 */
class NotTerm : public Formula {
private:
  Formula *operand;  // the operand

public:
  NotTerm(Formula *a)
    {
      operand = a;
    }
    
  virtual ~NotTerm()
    {
      delete operand;
    }

  virtual Formula* getOperand(int number)
    {
      return (number == 0) ? operand : (Formula*)NULL;
    }

  virtual int numberOfOperands()
    {
      return 1;
    }
    
  virtual char* operatorname()
    {
      return "NOT";
    }

  virtual void  print(ostream *out)
    {
      *out << operatorname() << " ";
      operand->print(out);
    }

  virtual void  prettyprint(ostream *out, int indentation = 0)
    {
      indent(out, indentation);
      *out << operatorname();
      operand->prettyprint(out, indentation+2);
    }
};

//****************************************************************

/* class for terms of the form  a,
   i.e., for variables. Variables are represented as strings.
   This is a leaf class of a COMPOSITE
 */
class VarTerm : public Formula {
private:
  char *name;

public:
  VarTerm(char *varname)
    {
      name = varname;
    }
    
  virtual Formula* getOperand(int number)
    {
      return (Formula*)NULL;
    }

  virtual int numberOfOperands()
    {
      return 0;
    }
    
  virtual char* operatorname()
    {
      return name;
    }

  virtual void  print(ostream *out)
    {
      *out << operatorname();
    }

  virtual void  prettyprint(ostream *out, int indentation = 0)
    {
      *out << operatorname();
    }
};

//****************************************************************

static void indent(ostream *out, int indentation)
    {
      *out << "\n";
      for (int i = 0; i < indentation; i++)
        *out << " ";
    }

//****************************************************************

int main(int argc, char**argv)
{
  Formula *term1 = new XorTerm(new VarTerm("a"), 
                               new NotTerm(new VarTerm("b")));

  NfoldTerm *term2 = new OrTerm();
  term2->addOperand(new NotTerm(new VarTerm("x1")));
  term2->addOperand(new NotTerm(new VarTerm("x2")));
  Formula *f = new AndTerm(term1, term2);
  f->print(&cout);   cout << "\n";
  f->prettyprint(&cout);
  return 0;
}

//****************************************************************
