
/* Communication channels program (COMO), C++
   Lutz Prechelt, Barbara Unger, 1997-11-10
   $Id: mo.cc,v 1.3 1997/11/10 18:04:40 prechelt Exp $
*/

#include "CommServices.h"

//************************************************************

/* class implementing a basic bidirectional communication channel with
   optional compression, encryption, and logging.
 */
class CommChannel {
public:
  enum State  { opened,  // able to transmit and receive
                closed,  // initial state
                failed   // opened, but had an error
              };
  enum Result { OK,         // operation was performed, state may change
                failure,    // operation failed, go to 'failed' state
                impossible  // operation was rejected, state is unchanged
              };

private:
  // general:
  State    channelState;
  // basic channel:
  char    *partner;
  Socket  *channel;           // the actual basic channel object
  // compression:
  bool     doCompression;
  // encryption:
  bool     doEncryption;
  EncryptionStream *encrypt;  // encryption state machine for sending
  DecryptionStream *decrypt;  // decryption state machine for receiving
  // logging:
  bool     doLogging;
  char    *logname; // name of this channel in the log file
  ostream *log;     // the log file

  // open the basic channel
  Result basicOpen();

  /* channels may be opened and closed multiple times. During each such cycle
     multiple transmits and receives are allowed. Channels are bidirectional.
   */
public:
  // create a basic communication channel (no extra functionality):
  CommChannel(char *partner);
  // create a communication channel which will have turned on
  // compression if compress == true,
  // encryption if key != NULL, and
  // logging if logchannelname != NULL
  CommChannel(char *partner, bool compress, char *key, char *logchannelname);
  virtual ~CommChannel();
  // open the channel (after creation or close):
  virtual Result open();
  // send some data over the channel:
  virtual Result transmit(Datablock *d);
  // receive data sent by 'transmit' from my partner channel:
  virtual Result receive(Datablock *d);
  // clear a 'failed' state if possible:
  virtual Result reset();
  // close the channel (after use):
  virtual Result close();
};

//************************************************************

// open the basic channel:
Result CommChannel::basicOpen()
  {
    if (channelState != closed)
      return impossible;
    if (channel == NULL)    // not yet initialized
      channel = CommServices::getPartner(partner);
    if (channel == NULL) {  // initialization failed
      channelState = failed;
      return failure;
    }
    else {                  // initialization OK, now open:
      if (channel->open() < 0) {  // open failed
        delete channel;
        channel = NULL;
        channelState = failed;
        return failure;
      }
      else {
        channelState = opened;
        return OK;
      }
    }
  }

//************************************************************

// create a basic communication channel (no extra functionality):
CommChannel(char *partner)
  {
    // general:
    channelState = closed;
    // basic channel:
    this->partner = partner;
    this->channel = NULL;
    // compression, encryption, logging:
    doCompression = doEncryption = doLogging = false;
  }

// create a communication channel which will have turned on
// compression if compress == true,
// encryption if key != NULL, and
// logging if logchannelname != NULL
CommChannel::CommChannel(char *partner, bool compress, char *key, 
                         char *logchannelname)
  {
    // general:
    channelState = closed;
    // basic channel:
    this->partner = partner;
    this->channel = NULL;
    // compression:
    doCompression = compress;
    // encryption:
    doEncryption = (key != NULL);
    if (doEncryption) {
      encrypt = new EncryptionStream(key);
      decrypt = new DecryptionStream(key);
    }
    // logging:
    doLogging = (logchannelname != NULL);
    logname = logchannelname;
  }

virtual CommChannel::~CommChannel()
  {
    // basic channel:
    delete partner;
    delete channel;
    // logging:
    if (doLogging) {
      *log << logname << ": deleted " << channelState << "\n";
      delete logname;
    }
    // encryption:
    if (doEncryption) {
      delete encrypt;
      delete decrypt;
    }
  }

//************************************************************

// open the channel (after creation or close):
virtual Result CommChannel::open()
  {
    Result result = basicOpen();
    // encryption, compression: 
    //   nothing to do
    // logging:
    if (doLogging) {
      log = Logging::theLogFile();
      *log << logname << ": open " << result
           << " " << channelState << "\n";
    }
    return result;
  }

// send some data over the channel:
virtual Result CommChannel::transmit(Datablock *d)
  {
    Result     result;
    Datablock *compressed,
              *encrypted;
    // compression:
    if (doCompression) {
      compressed = new Datablock((int)(0.8 * d->size()));
      Compressor::compress(d, compressed);
    }
    else
      compressed = d;  // do not compress, but do 'as if'
    // encryption:
    if (doEncryption) {
      encrypted = new Datablock(encrypt->encryptedsize(d->size()));
      encrypt->encrypt(compressed, encrypted);
    }
    else
      encrypted = compressed;  // do not encrypt, but do 'as if'
    // basic channel:
    if (channel->write(encrypted) < 0) {
      channelState = failed;
      result = failure;
    }
    else
      result = OK;
    // logging:
    if (doLogging) {
      *log << logname << ": T " << encrypted->size() << " "
           << encrypted->fingerprint() << " " << result << "\n";
    }
    // cleanup:
    if (doCompression)
      delete(compressed);
    if (doEncryption)
      delete(encrypted);
    return result;

  }

// receive data sent by 'transmit' from my partner channel:
virtual Result CommChannel::receive(Datablock *d)
  {
    Result     result;
    Datablock *encrypted,
              *compressed;
    // compression:
    if (doCompression)
      compressed = new Datablock();
    else
      compressed = d;
    // encryption:
    if (doEncryption)
      encrypted = new Datablock();
    else
      encrypted = compressed;
    // basic channel:
    if (channel->read(encrypted) < 0) {
      channelState = failed;
      result = failure;
    }
    else
      result = OK;
    // logging:
    if (doLogging) {
      *log << logname << ": R " << encrypted->size() << " "
           << encrypted->fingerprint() << " " << result << "\n";
    }
    // encryption:
    if (doEncryption) {
      decrypt->decrypt(encrypted, compressed);
      delete(encrypted);
    }
    // compression:
    if (doCompression) {
      int r2 = Compressor::decompress(compressed, d);
      delete(compressed);
      if (r2 < 0) {
        channelState = failed;
        result = failure;
      }
    }
    return result;
  }

// clear a 'failed' state if possible:
virtual Result CommChannel::reset()
  {
    State  oldstate = channelState;
    Result result;
    switch (channelState) {
    case opened:
      result = OK;
    case closed:
      result = impossible;
    case failed:  // recover from error if possible:
      // encryption:
      if (doEncryption) {
        // cannot reset an encrypted connection, because losing data
        // invalidates the internal state of the encryption:
        result = impossible;
      }
      // basic channel:
      else if (channel != NULL) {
        channel->clear();
        channelState = opened;
        result = OK;
      }
      else
        result = failure; // open failed: recovery is impossible
    }
    // logging:
    if (doLogging) {
      *log << logname << ": reset " << result
           << " " << oldstate << "\n";
    }
    return result;
  }

// close the channel (after use):
virtual Result CommChannel::close()
  {
    Result result;
    State  oldstate = channelState;
    switch (channelState) {
    case opened:  
    case failed:
      // basic channel:
      if (channel != NULL)
        channel->close();
      // encryption:
      if (doEncryption) {
        encrypt->reset();  // set back to initial state for next connection
        decrypt->reset();  // set back to initial state for next connection
      }
      channelState = closed;
      result = OK;
    case closed:
      result = impossible;
    }
    // logging:
    if (doLogging) {
      *log << logname << ": close " << result
           << " " << oldstate << "\n";
    }
    return result;
  }

//****************************************************************
