
/* Communication channels program (CODP), C++
   Barbara Unger, Lutz Prechelt, 1997-11-09
   $Id: pd.cc,v 1.2 1997/11/09 18:53:34 prechelt Exp $
*/

/* Inheritance tree:
   CommChannel
      BareChannel
      CommLayer
         SecuredChannel
         CompressorChannel
         LoggedChannel
*/

#include "CommServices.h"

//************************************************************

/* interface of all communication channels.
   root class of a DECORATOR.
 */
class CommChannel {
public:
  enum State  { opened,  // able to transmit and receive
                closed,  // initial state
                failed   // opened, but had an error
              };
  enum Result { OK,         // operation was performed, state may change
                failure,    // operation failed, go to 'failed' state
                impossible  // operation was rejected, state is unchanged
              };

  // channels may be opened and closed multiple times. During each such cycle
  // multiple transmits and receives are allowed. Channels are bidirectional.
  virtual State  state() = 0;    // current state of the channel
  virtual Result open() = 0;     // open channel (after creation or close)
  virtual Result transmit(Datablock *d) = 0; // send data over channel
  virtual Result receive(Datablock *d) = 0;  // receive data from channel
  virtual Result reset() = 0;    // try to clear a 'failed' state
  virtual Result close() = 0;    // close channel (after use)
  virtual void   fail() = 0;     // signal a failure [internal operation]
};

//************************************************************

/* basic communication channel, can be decorated (is part of a DECORATOR)
 */
class BareChannel : public CommChannel {
private:
  char    *partner;
  Socket  *channel;
  State    channelState;

  virtual void fail()
    {
      channelState = CommChannel::failed;
    }

public:
  BareChannel(char *partner)
    {
      this->partner = partner;
      this->channel = NULL;
      channelState = CommChannel::closed;
    }

  virtual ~BareChannel()
    {
      delete partner;
      delete channel;
    }

  virtual State  state()
    {
      return channelState;
    }

  virtual Result open()
    {
      if (channelState != closed)
        return CommChannel::impossible;
      if (channel == NULL)    // not yet initialized
        channel = CommServices::getPartner(partner);
      if (channel == NULL) {  // initialization failed
        channelState = CommChannel::failed;
        return CommChannel::failure;
      }
      else {                  // initialization OK, now open:
        if (channel->open() < 0) {  // open failed
          delete channel;
          channel = NULL;
          channelState = CommChannel::failed;
          return CommChannel::failure;
        }
        else {
          channelState = CommChannel::opened;
          return CommChannel::OK;
        }
      }
    }

  virtual Result transmit(Datablock *d)
    {
      if (channel->write(d) < 0) {
        channelState = CommChannel::failed;
        return CommChannel::failure;
      }
      else
        return OK;
    }

  virtual Result receive(Datablock *d)
    {
      if (channel->read(d) < 0) {
        channelState = CommChannel::failed;
        return CommChannel::failure;
      }
      else
        return OK;
    }

  virtual Result reset()
    {
      switch (channelState) {
      case opened:
        return CommChannel::OK;
      case closed:
        return CommChannel::impossible;
      case failed:  // recover from error if possible:
        if (channel != NULL) {
          channel->clear();
          channelState = CommChannel::opened;
          return CommChannel::OK;
        }
        else
          return CommChannel::failure; // open failed: recovery is impossible
      }
    }

  virtual Result close()
    {
      switch (channelState) {
      case opened:  
      case failed:
        if (channel != NULL)
	  channel->close();
        channelState = CommChannel::closed;
        return CommChannel::OK;
      case closed:
        return CommChannel::impossible;
      }
      
    }
};

//************************************************************

/* abstract decorator base class (part of a DECORATOR)
 */
class CommLayer : public CommChannel {
protected:
  CommChannel *baseChannel;  // the channel underlying this one

  virtual void fail()
    {
      baseChannel->fail();
    }

public:
  virtual State  state() 
    {
      return baseChannel->state();
    }

  virtual Result open() = 0;
  virtual Result transmit(Datablock *d) = 0;
  virtual Result receive(Datablock *d) = 0;
  virtual Result reset() = 0;
  virtual Result close() = 0;
};


//************************************************************

/* a DECORATOR class for compression
 */
class CompressorChannel : public CommLayer {
public:
  CompressorChannel(CommChannel *channel)
    {
      baseChannel = channel;
    }

  virtual Result open()
    {
      return baseChannel->open();
    }

  virtual Result transmit(Datablock *d)
    {
      Datablock *compressed = new Datablock((int)(0.8 * d->size()));
      Compressor::compress(d, compressed);
      Result r = baseChannel->transmit(compressed);
      delete(compressed);
      return r;
    }

  virtual Result receive(Datablock *d)
    {
      Datablock *compressed = new Datablock();
      Result r = baseChannel->receive(compressed);
      int r2 = Compressor::decompress(compressed, d);
      delete(compressed);
      if (r2 < 0) {
        fail();
        return CommChannel::failure;
      }
      else
        return r;
    }

  virtual Result reset()
    {
      return baseChannel->reset();
    }

  virtual Result close()
    {
      return baseChannel->close();
    }
};

//************************************************************

/* a DECORATOR class for encryption
 */
class SecuredChannel : public CommLayer {
private:
  EncryptionStream *encrypt;  // encryption state machine for sending
  DecryptionStream *decrypt;  // decryption state machine for receiving

public:
  SecuredChannel(char *key, CommChannel *channel)
    {
      baseChannel = channel;
      encrypt = new EncryptionStream(key);
      decrypt = new DecryptionStream(key);
    }

  virtual ~SecuredChannel()
    {
      delete encrypt;
      delete decrypt;
    }

  virtual Result open()
    {
      return baseChannel->open();
    }

  virtual Result transmit(Datablock *d)
    {
      Datablock *encrypted = new Datablock(encrypt->encryptedsize(d->size()));
      encrypt->encrypt(d, encrypted);
      Result r = baseChannel->transmit(encrypted);
      delete(encrypted);
      return r;
    }

  virtual Result receive(Datablock *d)
    {
      Datablock *encrypted = new Datablock();
      Result r = baseChannel->receive(encrypted);
      decrypt->decrypt(encrypted, d);
      delete(encrypted);
      return r;
    }

  virtual Result reset()
    {
      if (baseChannel->state() == CommChannel::failed)
        // cannot reset an encrypted connection, because losing data
        // invalidates the internal state of the encryption:
        return CommChannel::impossible;
      else
        // nothing to do here. Do whatever the underlying channel suggests:
        return baseChannel->reset();
    }

  virtual Result close()
    {
      encrypt->reset();  // set back to initial state for next connection
      decrypt->reset();  // set back to initial state for next connection
      return baseChannel->close();
    }
};


//************************************************************

/* a DECORATOR class for event logging
 */
class LoggedChannel : public CommLayer {
private:
  char     *name; // name of this channel in the log file
  ostream  *log;  // the log file
public:
  LoggedChannel(char *name, CommChannel *channel)
    {
      this->name = name;
      baseChannel = channel;
    }

  virtual ~LoggedChannel()
    {
      *log << name << ": deleted " << baseChannel->state() << "\n";
      delete name;
    }

  virtual Result open()
    {
      log = Logging::theLogFile();
      Result r = baseChannel->open();
      *log << name << ": open " << r 
	   << " " << baseChannel->state() << "\n";
      return r;
    }

  virtual Result transmit(Datablock *d)
    {
      Result r = baseChannel->transmit(d);
      *log << name << ": T " << d->size() << " "
	   << d->fingerprint() << " " << r << "\n";
      return r;
    }

  virtual Result receive(Datablock *d)
    {
      Result r = baseChannel->receive(d);
      *log << name << ": R " << d->size() << " "
	   << d->fingerprint() << " " << r << "\n";
      return r;
    }

  virtual Result reset()
    {
      State  oldstate = baseChannel->state();
      Result r = baseChannel->reset();
      *log << name << ": reset " << r 
	   << " " << oldstate << "\n";
      return r;
    }

  virtual Result close()
    {
      State  oldstate = baseChannel->state();
      Result r = baseChannel->close();
      *log << name << ": close " << r
	   << " " << oldstate << "\n";
      return r;
    }
};

//************************************************************
