
/* Graphics library program (GRMO), C++
   Barbara Unger, Lutz Prechelt, 1997-11-09
   $Id: both.cc,v 1.4 1997/11/10 18:04:32 prechelt Exp prechelt $
*/

/* class list (textual order):
 Point
 Context
 GraphicalObject
 Line
 Circle
 Objectgroup
 Generator

 Context_G
 Circle_G

 Context_A
 Circle_A

inheritance hierarchy:
 Point
 Context
    Context_G
    Context_A
 GraphicalObject
    Line
    Circle
       Circle_G
       Circle_A
 Objectgroup
 Generator

*/

#include <assert.h>
#include <math.h>
#include <stream.h>

#include "Window.h"

//************************************************************

/* representing and handling points in 2D space:
 */
class Point {
public:
  double x;
  double y;
  
  Point()
  {
    this->x = this->y = 0;
  }
    
  Point(double x, double y)
  {
    this->x = x;
    this->y = y;
  }

  // rotate a Point around an arbitrary point 'center' by an arbitrary
  // angle (in radiant, i.e., full circle is 2*M_PI):
  virtual void rotate(Point center, double angle)
  {
     double xnew, ynew;
     xnew =  cos(angle) * (x - center.x) -
 	     sin(angle) * (y - center.y) +
             center.x;
     ynew =  cos(angle) * (y - center.y) +
 	     sin(angle) * (x - center.x) +
             center.y;
     x = xnew;
     y = ynew;
  }

  // move a point by some distance:
  virtual void shift(double dx, double dy)
  {
    x += dx;
    y += dy;
  }

};

//************************************************************

/* interface for basic graphical operations and coordinate transformation
 */
class Context {
public:
  // generate context with default physical size and default
  // user coordinate system x,y in -100...100
  Context() {}
  // generate context with given physical size and user coordinate system
  Context(int phxsize, int physize,
           double xmin, double ymin, double xmax, double ymax) {}
  virtual void erase() = 0;           // erase graphical drawing area
  virtual double xResolution() = 0;    // pixels per x coordinate unit
  virtual double yResolution() = 0;    // pixels per y coordinate unit
  virtual void drawPoint(Point) = 0;  // draw a Point
  virtual void drawLine(Point, Point) = 0; // draw a Line
  // update display, i.e. make effects of drawX visible if necessary
  virtual void display() = 0;  // (will be null for some context types)
};

//************************************************************

/* Interface class for graphical objects
 *  (currently the Line and Circle classes)
   Each primitive graphical object will be implemented by one or
   maybe several subclasses:
   - one subclass S for the general part (in particular basic data
     representation and coordinate computations)
   - maybe one subsubclass of S for each type of graphical context
   The latter implement the 'draw' operation and allow for using more
   powerful primitives for certain context types if necessary.
   If the general 'Context' interface is sufficient, these classes are
   not present and 'draw' will be implemented in S.
 */
class GraphicalObject {
public:
  GraphicalObject() {}
  // rotate abstract representation of component around some point:
  virtual void rotate(Point center, double angle) = 0;
  // shift abstract representation of component by some distance:
  virtual void shift(double dx, double dy) = 0;
  // put a visible copy of the abstract object in its current state into k:
  virtual void draw(Context &k) = 0;
};

//************************************************************

/* implementation class for 'Line' graphical objects
 */
class Line : public GraphicalObject {
private:
  Point start;
  Point end;

public:
  Line() {}

  Line(Point start, Point end)
  {
    this->start = start;
    this->end   = end;
  }
  
  virtual void rotate(Point center, double angle)
  {
    start.rotate(center, angle);
    end.rotate(center, angle);
  }

  virtual void shift(double dx, double dy)
  {
    start.shift(dx, dy);
    end.shift(dx, dy);
  }

  virtual void draw(Context &k)
  {
    k.drawLine(start, end);
  }
};

//************************************************************

/* base class for 'Circle' graphical objects. 
   Will be further refined: subclasses will implement 'draw'
 */
class Circle : public GraphicalObject {
protected:
  Point center;
  double radius;

public:
  Circle() {}

  Circle(Point center, double radius)
  {
    this->center  = center;
    this->radius = radius;
  }
  
  virtual void rotate(Point center, double angle)
  {
    center.rotate(center, angle);
  }

  virtual void shift(double dx, double dy)
  {
    center.shift(dx, dy);
  }
};

//************************************************************

/* Container class for 'GraphicalObject' objects. 
   Implements operations on sets of GraphicalObjects
 */
class Objectgroup {
private:
  const int limit = 100;    // (group size limit is fixed)
  GraphicalObject *element[limit];
  int noOfElements;
  
public:
  Objectgroup()
  {
    noOfElements = 0;
  }
  
  /* inserts a GraphicalObject into an Objectgroup
  */
  virtual void add(GraphicalObject *k)
  {
    if (noOfElements < limit)
      element[noOfElements++] = k;
  }

  // inserts all GraphicalObjects of Objectgroup g into this Objectgroup:
  virtual void add(Objectgroup *g)
  {
    for (int i = 0;
         i < g->noOfElements && this->noOfElements < limit;
         i++)
      this->element[noOfElements++] = g->element[i];
  }

  // rotates all elements of Objectgroup:
  virtual void rotate (Point center, double angle)
  {
    for (int i = 0; i < noOfElements; i++)
      element[i]->rotate(center, angle);
  }

  // shifts all elements of Objectgroup:
  virtual void shift(double dx, double dy)
  {
    for (int i = 0; i < noOfElements; i++)
      element[i]->shift(dx, dy);
  }

  // draws all elements of Objectgroup:
  virtual void draw(Context &k)
  {
    for (int i = 0; i < noOfElements; i++)
      element[i]->draw(k);
  }
};

//************************************************************

/* Generator is a class containing factory methods for generating
   graphical objects according so one selected kind of context.
   Each method allocates, initializes, and returns an object of the
   respective kind.
 */
class Generator {
public:
  enum Generatortype { GRAPHICAL, ALPHANUMERICAL };
private:
  Generatortype type;

public:

  static Generator* newGenerator(enum Generatortype t); // see below
  Generator(Generatortype t) { type = t; }

  Context* newContext();
  Context* newContext(int gxsize, int gysize,
      double xmin, double ymin, double xmax, double ymax);
  Objectgroup* newObjectgroup();
  Line*    newLine(Point from, Point to);
  Circle*  newCircle(Point center, double radius);
};

//************************************************************
// classes for GRAPHICAL contexts:
//************************************************************

/* Context type for actual pixel graphics presentation in a windowing
   system. This class provides one operation beyond the standard 'Context'
   interface, namly drawCircle()
 */
class Context_G : public Context {
  int    gxsize, gysize;  // size of graphical drawing area (in pixels)
  double xmin, xmax;      // x axis of coordinate system
  double ymin, ymax;      // y axis of coordinate system
  Window *window;         // actual implementation of drawing area

  /* transform user-coordinate point p into device coordinates x, y
     and return whether the point will be visible
  */
  virtual bool transform(Point p, int *x, int *y)
  {
    double xfraction = (p.x - xmin) / (xmax - xmin),
           yfraction = (p.y - ymin) / (ymax - ymin);
    *x = (int)((gxsize-1) * xfraction + 0.5);
    *y = (int)((gysize-1) * yfraction + 0.5);
    return (*x >= 0 && *x < gxsize && *y >= 0 && *y < gysize);
  }

public:
  Context_G()
  {
    gxsize = gysize = 201; // default size
    window = Window::createFrame("Graphics", gxsize, gysize);
    xmin = ymin = -100.0;
    xmax = ymax =  100.0;
  }

  Context_G(int gxsize, int gysize,
            double xmin, double ymin, double xmax, double ymax)
  {
    this->gxsize = gxsize;
    this->gysize = gysize;
    window = Window::createFrame("Graphics", gxsize, gysize);
    this->xmin = xmin;
    this->xmax = xmax;
    this->ymin = ymin;
    this->ymax = ymax;
  }

  virtual ~Context_G()
  {
    delete window;
  }

  virtual double xResolution()
  {
    return gxsize / (xmax - xmin);
  }
  
  virtual double yResolution()
  {
    return gysize / (ymax - ymin);
  }

  virtual void erase()
  {
    window->clear();
  }

  virtual void drawPoint(Point p)
  {
    int x, y;
    if (transform(p, &x, &y))
      window->drawPoint(x, y);
  }

  virtual void drawLine(Point from, Point to)
  {
    int x0, y0, x1, y1;
    transform(from,  &x0, &y0);
    transform(to, &x1, &y1);
    window->drawLine(x0, y0, x1, y1);
  }

  virtual void display()
  {
    // no action required, everything is already visible
  }

  /* EXTENSION: draw a circle of a given radius (expressed in x-coordinate
     units)
   */
  virtual void drawCircle(Point center, double radius)
  {
    int x, y, r;
    transform(center,  &x, &y);
    // radius is converted in terms of x-coordinates:
    r = (int)(radius / (xmax-xmin) * gxsize);
    window->drawCircle(x, y, r);
  }

};

//************************************************************

/* implementation class for 'Circle' graphical objects in GRAPHICAL contexts:
   utilize the primitive circle drawing operation of this context type.
 */
class Circle_G : public Circle {

public:
  Circle_G() {}

  Circle_G(Point center, double radius) : Circle(center, radius) {}
  
  virtual void draw(Context &k)
  {
    // just use the extended interface of Context_G:
    ((Context_G&)k).drawCircle(center, radius);
  }
};

//************************************************************
// classes for ALPHANUMERICAL contexts:
//************************************************************

/* context type for character-based pseudo-graphics on alphanumerical
   displays
 */
class Context_A : public Context {
  int    gxsize, gysize; // size of drawing area (in characters)
  char   *drawingarea;   // two dimensional array: graphical contents
  double xmin, xmax;     // x axis of coordinate system
  double ymin, ymax;     // y axis of coordinate system
  
private:
  // auxiliary function: access point (x,y) in drawing area:
  char& c(int x, int y)
  {
    assert (x >= 0 && x < gxsize && y >= 0 && y < gysize);
    return drawingarea[x + gxsize*y];
  }

public:
  Context_A()
  {
    gxsize = 80;
    gysize = 24;
    drawingarea = new char[gxsize*gysize];
    xmin = ymin = -100.0;
    xmax = ymax =  100.0;
    erase();
  }

  Context_A(int gxsize, int gysize,
             double xmin, double ymin, double xmax, double ymax)
  {
    this->gxsize = gxsize;
    this->gysize = gysize;
    drawingarea = new char[gxsize*gysize];
    this->xmin = xmin;
    this->xmax = xmax;
    this->ymin = ymin;
    this->ymax = ymax;
    erase();
  }

  virtual ~Context_A()
  {
    delete drawingarea;
  }

  virtual double xResolution()
  {
    return gxsize / (xmax - xmin);
  }
  
  virtual double yResolution()
  {
    return gysize / (ymax - ymin);
  }

  virtual void erase()
  {
    // clear all contents:
    for (int x = 0; x < gxsize; x++)
      for (int y = 0; y < gysize; y++)
        c(x, y) = ' ';
    // paint a border as an optical marker:
    for (int x = 0; x < gxsize; x++)
      c(x, 0) = c(x, gysize-1) = '-';
    for (int y = 0; y < gysize; y++)
      c(0, y) = c(gxsize-1, y) = '|';
    c(0, 0)        = c(0, gysize-1)        = '+';
    c(gxsize-1, 0) = c(gxsize-1, gysize-1) = '+';
  }

  virtual void drawPoint(Point p)
  {
    double xfraction = (p.x - xmin) / (xmax - xmin),
           yfraction = (p.y - ymin) / (ymax - ymin);
    int    x = (int)((gxsize-1) * xfraction + 0.5),
           y = (int)((gysize-1) * yfraction + 0.5);
    if (x >= 0 && x < gxsize && y >= 0 && y < gysize) // if Point is visible
      c(x, y) = '#';
  }

  virtual void drawLine(Point from, Point to)
  {
    double xpixel = fabs(from.x - to.x) * xResolution(); // number of pixels
    double ypixel = fabs(from.y - to.y) * yResolution(); // number of pixels
    // we draw pointwise. Therefore, the interpolation steps must not be
    // larger than 1 pixel equivalent:
    double noOfSteps = (xpixel > ypixel ? xpixel : ypixel);
    double xstep = (to.x - from.x) / noOfSteps;
    double ystep = (to.y - from.y) / noOfSteps;
    Point  p = from;
    // now draw the line:
    drawPoint(from);
    for (int i = 1; i <= (int)noOfSteps; i++) {
      drawPoint(p);
      p.x += xstep;
      p.y += ystep;
    }
    drawPoint(to);
  }

  virtual void display()
  {
    /* any draw operation works on the internal buffer 'drawingarea' only.
       Hence we have to display this buffer now:
    */
    cout << "\n";
    for (int y = gysize-1; y >= 0; y--) {
      for (int x = 0; x < gxsize; x++)
        cout << c(x, y);
      cout << "\n";
    }
  }
};

//************************************************************

/* implementation class for 'Circle' graphical objects in ALPHANUMERICAL
   contexts: drawing operation needs to be implemented 'by hand'.
 */
class Circle_A : public Circle {

public:
  Circle_A() {}

  Circle_A(Point center, double radius) : Circle(center, radius) {}
  
  virtual void draw(Context &k)
  {
    Point from, to = Point(center.x + radius, center.y);
    int totalpixels = (int)(radius * 2 * M_PI * 
                            (k.xResolution() + k.yResolution())/2);
    double angle = 0;
    double anglestep = 2 * M_PI / (totalpixels > 100 ? 100 : totalpixels);
    for (int i = 0; i < totalpixels; i++) {
      from = to;
      angle += anglestep;
      to.x = center.x + cos(angle)*radius;
      to.y = center.y + sin(angle)*radius;
      k.drawLine(from, to);
    }
  }
};

//************************************************************
//************************************************************

Generator* Generator::newGenerator(enum Generatortype e)
{
  return new Generator(e);
}

Context* Generator::newContext()
{
  switch (type) {
  case ALPHANUMERICAL:  return new Context_A();
  case GRAPHICAL:       return new Context_G();
  default:              return NULL;
  }
}

Context* Generator::newContext(int gxsize, int gysize,
      double xmin, double ymin, double xmax, double ymax)
{
  switch (type) {
  case ALPHANUMERICAL:  return new Context_A(gxsize, gysize, 
                                             xmin, ymin, xmax, ymax);
  case GRAPHICAL:       return new Context_G(gxsize, gysize, 
                                             xmin, ymin, xmax, ymax);
  default:              return NULL;
  }
}

Objectgroup* Generator::newObjectgroup()
{
  //Objectgroup is always independent of Context type:
  return new Objectgroup();
}

Line* Generator::newLine(Point from, Point to)
{
  // Line _happens_ to be independent of Context type:
  switch (type) {
  case ALPHANUMERICAL:
  case GRAPHICAL:       return new Line(from, to);
  default:              return NULL;
  }
}

Circle* Generator::newCircle(Point center, double radius)
{
  switch (type) {
  case ALPHANUMERICAL:  return new Circle_A(center, radius);
  case GRAPHICAL:       return new Circle_G(center, radius);
  default:              return NULL;
  }
}

//************************************************************

void cross(Generator *generator)
{
  Objectgroup *pair = generator->newObjectgroup();
  Objectgroup *cross = generator->newObjectgroup();
  // generate a pair of vertical lines
  pair->add(generator->newLine(Point(-2, -10), Point(-2, 10)));
  pair->add(generator->newLine(Point( 2, -10), Point( 2, 10)));
  // tilt pair to the right and add it to 'cross':
  pair->rotate(Point(0, 0), 45 * (2*M_PI/360));
  cross->add(pair);
  // tilt pair to the left and add it to 'cross':
  pair->rotate(Point(0, 0), -90 * (2*M_PI/360));
  cross->add(pair);
  // draw cross:
  cross->draw(*(generator->newContext()));
}

//************************************************************

int main(int argc, char **argv)
{
  Generator     *generator = 
      Generator::newGenerator(Generator::ALPHANUMERICAL);
  Context      *k = generator->newContext(78, 24, 0, 0, 150, 100);
  Objectgroup  *olympia = generator->newObjectgroup();
  Line         *line;
  Point         p;
  double        radius = 25;
  double        shift = 1.5 * radius;
  double        x0 = 35;
  double        y0 = 30;

  // generate Olympic logo (won't look very nice, though...):
  p.x = x0;
  p.y = y0;
  radius = 25;
  for (int i = 0; i < 3; i++) {
    olympia->add(generator->newCircle(p, radius));
    p.x += shift;
  }
  p.x = x0 + shift/2;
  p.y = y0 + radius;
  for (int i = 0; i < 2; i++) {
    olympia->add(generator->newCircle(p, radius));
    p.x += shift;
  }
  // so far olympia is upside down. Rotate it by 180 degrees:
  olympia->rotate(Point(x0 + shift, y0 + radius/2), M_PI);
  // add a top line:
  line = generator->newLine(Point(20, 90), Point(130, 90));

  // draw and show picture:
  k->erase();
  olympia->draw(*k);
  line->draw(*k);
  k->display();
  
  return 0;
}
