
#pd#
/* Stock ticker program (STDP), C++
#end#
#mo#
/* Stock ticker program (STMO), C++
#end#
   Walter F. Tichy, Lutz Prechelt, 1997-11-09
   $Id: both.cc,v 1.4 1997/11/10 18:04:43 prechelt Exp $
*/

/* class list and inheritance hierarchy:
   StockTrade
   TradeStream
      FakeTradeStream
   StockTicker
      SimpleStockTicker
      VolumeStockTicker
   TradeInfoSource
*/

#include <fstream.h>
#include <stdio.h>
#include <stream.h>
#include <string.h>
#include <unistd.h>

#include "Window.h"  // class Window

/* class for representing a single stock trading transaction
 */
class StockTrade {
public:
  char *symbol;  // short name of the stock
  char *price;   // price per share of trade
  char *volume;  // number of shares traded

  StockTrade()
    {
      symbol = price = volume = NULL;
    }

  StockTrade(char *symbol, char *price, char *volume)
    {
      // strings must be on heap, constructor takes ownership of them
      this->symbol = symbol;
      this->price  = price;
      this->volume = volume;
    }

  virtual ~StockTrade()
    {
      delete symbol;
      delete price;
      delete volume;
    }
};

/********************************************************************/

/* interface of all data sources for stock trade events
 */
class TradeStream {
public:

  /* create a new ticker stream reading from a given source: */
  TradeStream(char *stockExchangeName) {}

  /* return the next stock trade: */
  virtual StockTrade getTrade() = 0;
};

/********************************************************************/

/* class implementing a source of simulated stock trading events
 */
class FakeTradeStream : public TradeStream{
private:
    ifstream *in;

public:
  /* create a new ticker stream: */
  FakeTradeStream(char *stockExchangeName) :
    TradeStream(stockExchangeName)
    {
      in = new ifstream(stockExchangeName);
      if (in->fail()) {
        perror("can't open trading data file");
        exit(1);
      }
    }

  /* destroy ticker stream: */
  virtual ~FakeTradeStream()
    {
      in->close();
      delete in;
    }

  /* return the next stock trade: */
  virtual StockTrade getTrade() 
    {
      /* NOTE: This is just a simple simulation of a real stream of stock
         trading information. It reads records from a file (and starts
         over when the file is exhausted). In reality this should contact
         some online information service to receive its information.
      */
      char *sym = new char[20];
      char *price = new char[10];
      char *vol = new char[10];
      in->getline(sym, 20, ' ');
      if (in->eof() || in->fail()) {  // if file was read to end, start again
        in->clear();
        in->seekg(0, ios::beg);
        in->getline(sym, 20, ' ');
      }
      in->getline(price, 10, ' ');
      in->getline(vol,   10, '\n');
      sleep(1);  // artificial delay between two trade infos
      return StockTrade(sym, price, vol);
    }
};

/********************************************************************/

#pd#
class TradeInfoSource;  // TradeInfoSource and StockTicker use each other

/* interface for all stock ticker OBSERVER classes:
 */
class StockTicker {
public:
  virtual void announceNextTrade(TradeInfoSource *src) = 0;
#end#
#mo#
/* interface for all stock ticker display classes:
 */
class StockTicker {
public:
#end#
  virtual void showNextTrade(StockTrade trade) = 0;
};

/********************************************************************/

#pd#
/* concrete stock ticker OBSERVER class
#end#
#mo#
/* concrete stock ticker display class
#end#
 */
class SimpleStockTicker : public StockTicker {
private:
  int     columns;     // how many characters wide is the display?
  char   *contents;    // which characters are on the display?
  Window *window;      // window used for display

public:
  SimpleStockTicker(int columns);
  virtual ~SimpleStockTicker();
#pd#
  virtual void announceNextTrade(TradeInfoSource *src);
#end#
  virtual void showNextTrade(StockTrade trade);
};

/********************************************************************/

#pd#
/* concrete stock ticker OBSERVER class
#end#
#mo#
/* concrete stock ticker display class
#end#
 */
class VolumeStockTicker : public StockTicker {
private:
  Window *window;   // window used for display

public:
  VolumeStockTicker(int width);
  virtual ~VolumeStockTicker();
#pd#
  virtual void announceNextTrade(TradeInfoSource *src);
#end#
  virtual void showNextTrade(StockTrade trade);
};

/********************************************************************/

#pd#
/* subject class of OBSERVER: fetches data and updates the observers
#end#
#mo#
/* main class: fetches data and updates the displays
#end#
 */
class TradeInfoSource {
private:
  StockTrade   lastTrade;
  TradeStream *source;
#pd#
  StockTicker *display[100];
  int          displayCount;
#end#
#mo#
  StockTicker *simple;
#end#

public:
  TradeInfoSource(char *stockExchangeName)
    {
#pd#
      displayCount = 0;
#end#
      source = new FakeTradeStream(stockExchangeName);
#pd#
      subscribe(new SimpleStockTicker(60));
#end#
#mo#
      simple = new SimpleStockTicker(60);
#end#
      showTradeInfo();
    }

  virtual ~TradeInfoSource()
    {
#pd#
      for (int i = 0; i < displayCount; i++)
        delete display[i];
#end#
      delete source;
#mo#
      delete simple;
#end#
    }

#pd#
  virtual StockTrade latestTrade()
    {
      return lastTrade;
    }

  virtual int subscribe(StockTicker *s)
    {
      if (displayCount < 100)
        display[displayCount++] = s;
      return displayCount;
    }

  virtual int unsubscribe(StockTicker *s)
    {
      // removes s if that is subscribed:
      int pos;
      // search for s:
      for (pos = 0; pos < displayCount; pos++)
        if (display[pos] == s)
          break;
      // if not found:
      if (display[pos] != s)
        return displayCount;  // s is not subscribed
      // if found:
      displayCount--;
      for ( ; pos < displayCount; pos++)  // shift remaining ones to the front
        display[pos] = display[pos+1];
      return displayCount;
    }

  virtual int noOfDisplays()
    {
      return displayCount;
    }
    
  virtual StockTicker* getDisplay(int number)
    {
      if (number < 0 || number >= displayCount)
        return NULL;
      else
        display[number];
    }
#end#

  virtual void showTradeInfo()
    {
      for (;;) {
        lastTrade = source->getTrade();
#pd#
        for (int i = 0; i < displayCount; i++) {
          display[i]->announceNextTrade(this);
        }
#end#
#mo#
        simple->showNextTrade(lastTrade);
#end#
      }
    }


};

/********************************************************************/
/* implementations of methods:
/********************************************************************/

SimpleStockTicker::SimpleStockTicker(int columns)
    {
      this->columns = columns;
      contents = new char[columns+100];
      strcpy(contents, "");  // ticker contents are initially empty
      // create window representation:
      window = Window::createFrame("Simple Stock Ticker", 0, 0);
      window->setFont("Helvetica", 10);
      window->resize(window->fontWidth() * columns,
                    window->fontHeight() * 2);
    }

SimpleStockTicker::~SimpleStockTicker()
    {
      delete contents;
      delete window;
    }

#pd#
void SimpleStockTicker::announceNextTrade(TradeInfoSource *src)
    {
      StockTrade trade = src->latestTrade();
      showNextTrade(trade);
    }

#end#
void SimpleStockTicker::showNextTrade(StockTrade trade)
    {
      // add new trade info to contents:
      strcat(contents, trade.symbol);
      strcat(contents, " ");
      strcat(contents, trade.price);
      strcat(contents, " ** ");
      // remove old part that becomes invisible:
      int removepos = strlen(contents) - columns;
      strcpy(contents, contents + (removepos < 0 ? 0 : removepos));
      // update display:
      window->clear();
      window->drawString(contents, 
                        0, (int)(window->fontHeight() * 0.5));
    }

/********************************************************************/

VolumeStockTicker::VolumeStockTicker(int width)
    {
      window = Window::createFrame("Volume Stock Ticker", 0, 0,
                                   true); // window will store its contents
      window->setFont("Helvetica", 8);
      window->resize(width, window->fontHeight("1234567890") * 3 + 2);
      window->setColor(Window::black, Window::gray1);
    }

VolumeStockTicker::~VolumeStockTicker()
    {
      delete window;
    }

#pd#
void VolumeStockTicker::announceNextTrade(TradeInfoSource *src)
    {
      StockTrade trade = src->latestTrade();
      showNextTrade(trade);
    }

#end#
void VolumeStockTicker::showNextTrade(StockTrade trade)
    {
      // determine presentation size of new trade info:
      int symbolWidth;   // pixels needed for name
      int priceWidth;    // pixels needed for price
      int volumeWidth;   // pixels needed for volume
      int numbersWidth;  // maximum of priceWidth and volumeWidth
      window->fontAttributes(Window::bold, true);
      symbolWidth  = window->fontWidth(trade.symbol);
      window->fontAttributes(Window::bold, false);
      priceWidth   = window->fontWidth(trade.price);
      volumeWidth  = window->fontWidth(trade.volume);
      numbersWidth = priceWidth > volumeWidth ? priceWidth : volumeWidth;
      // shift window contents left as needed for new entry on the right:
      window->shiftContents(-window->fontWidth(" ") - symbolWidth - 2
                           - numbersWidth, 0); 
      // make new entry on the right, font is currently non-bold:
      window->drawString(trade.price,
                        window->width() - numbersWidth,
                        (int)(window->fontHeight() * 0.5 + 
                              window->fontHeight(trade.volume) + 2));
      window->drawString(trade.volume,
                        window->width() - numbersWidth,
                        (int)(window->fontHeight() * 0.5));
      window->fontAttributes(Window::bold, true);
      window->drawString(trade.symbol,
          window->width() - numbersWidth - 2 - symbolWidth,
          (int)(window->height()/2 - window->fontHeight(trade.symbol)/2));
      window->fontAttributes(Window::bold, false);
    }

/********************************************************************/
